import numpy as np
from .core import StatelessLayer, StatefulLayer


class Add(StatelessLayer):
    """Add up arrays element-wise."""
    def transform(self, *arrays):
        return np.sum(arrays, axis=0)


class Subtract(StatelessLayer):
    """Subtract one array from another."""
    def transform(self, X1, X2):
        return X1 - X2


class Multiply(StatelessLayer):
    """Multiply array by a given scalar.

    Parameters
    ----------
    factor : float
        multiplier.
    """
    def __init__(self, factor):
        super().__init__(factor=factor)

    def transform(self, X):
        return self.kwargs['factor'] * X


class Divide(StatelessLayer):
    """Divide given array by another given array element-wise.

    Parameters
    ----------
    impute : int/float or None
        the value to impute when encountering a divide by zero.
    """
    def __init__(self, impute=0):
        super().__init__(impute=impute)

    def transform(self, X1, X2):
        impute_array = np.ones_like(X1) * self.kwargs['impute']
        return np.divide(X1.astype(np.float16), X2, out=impute_array.astype(np.float16), where=X2!=0)


class Dot(StatelessLayer):
    """Multiply array by a given matrix, as matrix mulitplication.

    Parameters
    ----------
    W : np.array
        matrix by which to multiply.
    """
    def __init__(self, W):
        super().__init__(W=W)

    def transform(self, X):
        return np.dot(X, self.kwargs['W'])


class Normalize(StatelessLayer):
    """Divide array by total to cause it to sum to one. If zero array, make uniform."""
    def transform(self, X):
        out = X.copy()
        S = out.sum(axis=1)
        out[S<0.0001, :] = np.ones(out.shape[1])
        return out / out.sum(axis=1, keepdims=True)



import math
import numpy as np
import pandas as pd
from ..utils.generic import listify


class PandasGenerator:
    """A generator of data batches from a Pandas Dataframe into Megatron Input nodes.

    Parameters
    ----------
    dataframe : Pandas.DataFrame
        dataframe to load data from.
    batch_size : int
        number of observations to yield in each iteration.
    exclude_cols : list of str (default: [])
        any columns that should not be loaded as Input.
    """
    def __init__(self, dataframe, batch_size, exclude_cols=[]):
        self.batch_size = batch_size
        self.dataframe = dataframe
        self.n = 0
        if self.batch_size:
            self.n_batches = math.ceil(self.dataframe.shape[0] / self.batch_size)
        else:
            self.n_batches = 1
        self.exclude_cols = exclude_cols

    def __iter__(self):
        return self

    def __next__(self):
        if self.n == self.n_batches:
            self.n = 0

        if self.batch_size:
            start = self.n * self.batch_size
            end = min([self.dataframe.shape[0], start + self.batch_size])
            out = self.dataframe.iloc[start:end].drop(self.exclude_cols, axis=1)
        else:
            out = self.dataframe.drop(self.exclude_cols, axis=1)
        self.n += 1

        return dict(zip(out.columns, out.values.T))


class CSVGenerator:
    """A generator of data batches from a CSV file in pipeline Input format.

    Parameters
    ----------
    filepath : str
        the CSV filepath to be loaded from.
    batch_size : int
        number of observations to yield in each iteration.
    exclude_cols : list of str (default: [])
        any columns that should not be loaded as Input.
    """
    def __init__(self, filepath, batch_size, exclude_cols=[]):
        self.filepath = filepath
        self.batch_size = batch_size
        # take advantage of Pandas read_csv function to make it simpler and more robust
        if self.batch_size:
            self.cursor = pd.read_csv(self.filepath, chunksize=self.batch_size)
        else:
            self.cursor = self._make_generator()
        self.exclude_cols = exclude_cols

    def _make_generator(self):
        yield pd.read_csv(self.filepath)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            out = next(self.cursor).drop(self.exclude_cols, axis=1)
        except StopIteration:
            self.cursor = pd.read_csv(self.filepath, chunksize=self.batch_size)
        return dict(zip(out.columns, out.values.T))


class SQLGenerator:
    """A generator of data batches from a SQL query in pipeline Input format.

    Parameters
    ----------
    connection : Connection
        a database connection to any valid SQL database engine.
    query : str
        a valid SQL query according to the engine being used, that extracts the data for Inputs.
    batch_size : int
        number of observations to yield in each iteration.
    limit : int
        number of observations to use from the query in total.
    """
    def __init__(self, connection, query, batch_size, limit=None):
        self.batch_size = batch_size
        self.connection = connection
        self.query = query
        if limit:
            self.query += ' LIMIT {}'.format(nrows)
        self.cursor = self.connection.execute(self.query)
        self.names = [col[0] for col in self.cursor.description]

    def __iter__(self):
        return self

    def __next__(self):
        try:
            out = self.cursor.fetchmany(self.batch_size)
        except StopIteration:
            self.cursor = self.connection.execute(self.query)
            out = self.cursor.fetchmany(self.batch_size)
        coldata = np.array(out).T

        return dict(zip(self.names, coldata))



import re
import sqlite3
import pickle
import numpy as np
import pandas as pd
from .. import utils


class DataStore:
    """SQL table of input data and output features, associated with a single pipeline.

    Parameters
    ----------
    table_name : str
        name of pipeline's cache table in the database.
    version : str
        version tag for pipeline's cache table in the database.
    db_conn : Connection
        database connection to query.
    """
    def __init__(self, table_name, version, db_conn, overwrite):
        self.db = db_conn
        if version:
            self.table_name = '{}_{}'.format(table_name, version)
        else:
            self.table_name = table_name
        if overwite:
            self.db.execute("DROP TABLE IF EXISTS {}".format(self.table_name))

    def _check_schema(self, output_data):
        """If existing SQL colnames and data colnames are off, throw error that table is in use.

        Parameters
        ----------
        output_data : dict of ndarray
            resulting features from applying pipeline to input_data.
        """
        sql_cols = "SELECT sql FROM sqlite_master WHERE name='{}';"
        sql_cols = sql_cols.format(self.table_name)
        sql_cols = self.db.execute(sql_cols).fetchone()[0]
        sql_cols = np.array([c[1:-1] for c in re.findall('"[^\s]+"', sql_cols)])
        out_cols = np.insert(output_data.columns.values, 0, 'ind')
        if not np.array_equal(sql_cols, out_cols):
            msg = "Pipeline name already in use with different outputs: currently {}, not {}."
            msg = msg.format(sql_cols.tolist(), output_data.columns.values.tolist())
            raise ValueError(msg)

    def write(self, output_data, data_index):
        """Write set of observations to database.

        For features that are multi-dimensional, use pickle to compress to string.

        Parameters
        ----------
        output_data : dict of ndarray
            resulting features from applying pipeline to input_data.
        data_index : np.array
            index of observations.
        """
        self.dtypes = []
        self.original_shapes = []
        output_df = pd.DataFrame()

        # identify and pickle any multi-dimensional features
        for i, out_data in enumerate(output_data):
            self.original_shapes.append(out_data.shape[1:])
            if len(out_data.shape) > 1:
                self.dtypes.append('TEXT')
                flat_data = out_data.reshape((out_data.shape[0], -1))
                flat_data = np.apply_along_axis(lambda x: pickle.dumps(x), axis=1, arr=flat_data)
                output_df['output{}'.format(i)] = flat_data
            else:
                self.dtypes.append('REAL')
                output_df['output{}'.format(i)] = out_data

        # create table if it doesn't yet exist; if it does, check the schema matches the new data
        exists_query = 'SELECT name FROM sqlite_master WHERE type="table" AND name="{}";'
        exists_query = exists_query.format(self.table_name)
        if self.db.execute(exists_query).fetchone():
            self._check_schema(output_df)
        else:
            cols = ', '.join(['"output{}" {}'.format(i, dtype)
                              for i, dtype in enumerate(self.dtypes)])
            make_query = 'CREATE TABLE {} ("ind" VARCHAR, {});'
            self.db.execute(make_query.format(self.table_name, cols))
            self.db.commit()

        # delete data for indices that are being written
        inds = ','.join([str(i) for i in data_index.tolist()])
        self.db.execute("DELETE FROM {} WHERE ind IN ({})".format(self.table_name, inds))

        # write new data to db
        output_df['ind'] = data_index
        output_df.to_sql(self.table_name, self.db, if_exists='append', index=False)
        self.db.execute("CREATE UNIQUE INDEX IF NOT EXISTS ind ON {} (ind)".format(self.table_name))
        self.db.commit()

    def read(self, cols=None, rows=None):
        """Retrieve all processed features from cache, or lookup a single observation.

        For features that are multi-dimensional, use pickle to read string.

        Parameters
        ----------
        cols : list of int (default: None)
            indices of output columns to retrieve. If None, get all columns.
        rows: list of any or any (default: None)
            index value to lookup output for, in dictionary form. If None, get all rows.
            should be the same data type as the index.
        """
        if cols:
            cols = ['output{}'.format(c) for c in utils.generic.listify(cols)]
            cols = ', '.join(cols)
            query = "SELECT {}, ind FROM {} ".format(cols, self.table_name)
        else:
            query = "SELECT * FROM {} ".format(self.table_name)

        if rows:
            rows = utils.generic.listify(rows)
            query += "WHERE ind IN ({})".format(','.join(rows))

        out = list(pd.read_sql_query(query, self.db, index_col='ind').values.T)
        for i in range(len(out)):
            if self.dtypes[i] == 'TEXT':
                out[i] = np.array([pickle.loads(v) for v in out[i]])
                out[i] = out[i].reshape([-1] + list(self.original_shapes[i]))
        return out



import numpy as np
from ..nodes.auxiliary import MetricNode
from .. import utils


class Metric:
    """Layer type that holds an evaluation metric; only incorporated for Pipeline evaluation.

    Parameters
    ----------
    metric_fn : function
        the metric function to be wrapped.
    **kwargs
        any keyword arguments to be passed to the metric when being called.
    """
    def __init__(self, metric_fn, **kwargs):
        self.metric = metric_fn
        self.kwargs = kwargs

    def _call(self, nodes, name):
        """Produce a MetricNode acting on the given inbound nodes.

        Parameters
        ----------
        nodes : list of megatron.Node
            nodes to be passed to the metric function.
        name : str
            name to be associated with this MetricNode.
        """
        nodes = utils.generic.listify(nodes)
        out_node = MetricNode(self, nodes, name)
        for node in nodes:
            node.outbound_nodes.append(out_node)
        if all(node.output is not None for node in nodes):
            out_node.evaluate()
        return out_node

    def __call__(self, nodes, name):
        return self._call(nodes, name)

    def evaluate(self, *inputs):
        """Run metric function on given input data."""
        return self.metric(*inputs, **self.kwargs)



import numpy as np
from .core import StatelessLayer
try:
    import skimage
except ImportError:
    pass


class RGBtoGrey(StatelessLayer):
    """Convert an RGB array representation of an image to greyscale.

    Parameters
    ----------
    method : {'luminosity', 'lightness', 'average'}
    """
    def __init__(self, method='luminosity', keep_dim=False):
        super().__init__(method=method, keep_dim=keep_dim)

    def transform(self, X):
        if self.kwargs['method'] == 'lightness':
            out = (X.max(axis=2) + X.min(axis=2)) / 2.
        elif self.kwargs['method'] == 'average':
            out = X.mean(axis=2)
        elif self.kwargs['method'] == 'luminosity':
            out = np.matmul(X, np.array([0.21, 0.72, 0.07]))
        else:
            raise ValueError("Invalid averaging method for rgb_to_grey.")

        if self.kwargs['keep_dim']:
            out = np.expand_dims(out, -1)
        return out


class RGBtoBinary(StatelessLayer):
    """Convert image to binary mask where a 1 indicates a non-black cell."""
    def transform(self, X):
        return (X.max(axis=2) > 0).astype(int)[:, :, np.newaxis]


class Downsample(StatelessLayer):
    """Shrink an image to a given size proportionally.

    Parameters
    ----------
    new_shape : tuple of int
        the target shape for the new image.
    """
    def __init__(self, new_shape):
        super().__init__(new_shape=new_shape)

    def transform(self, X):
        if any(self.kwargs['new_shape'][i] > X.shape[i]
               for i in range(len(self.kwargs['new_shape']))):
            raise ValueError("New shape is larger than current in at least one dimension.")
        return skimage.transform.resize(X, self.new_shape)


class Upsample(StatelessLayer):
    """Expand an image to a given size proportionally.

    Parameters
    ----------
    new_shape : tuple of int
        the target shape for the new image.
    """
    def __init__(self, new_shape):
        super().__init__(new_shape=new_shape)

    def transform(self, X):
        if any(self.kwargs['new_shape'][i] < X.shape[i]
               for i in range(len(self.kwargs['new_shape']))):
            raise ValueError("New shape is smaller than current in at least one dimension.")
        return skimage.transform.resize(X, self.kwargs['new_shape'])



import numpy as np
import pandas as pd
from .core import InputNode


def from_dataframe(df, exclude_cols=[], eager=False, nrows=None):
    """Load Input nodes from columns of a Pandas dataframe.

    Parameters
    ----------
    df : Pandas.DataFrame
        dataframe from which to load columns.
    exclude_cols : list of str (default: [])
        any columns that should not be loaded as Input.
    eager : bool
        whether to load data as well, making for eager execution.
    nrows : int
        number of rows to load when eager is True. Default is for all rows to load.
    """
    exclude_cols = set(exclude_cols)
    cols = [col for col in df.columns if not col in exclude_cols]
    nodes = [InputNode(col) for col in cols]
    if eager:
        nodes = [node(df[col].values[:nrows]) for node, col in zip(nodes, cols)]
    return {node.name: node for node in nodes}


def from_csv(filepath, exclude_cols=[], eager=False, nrows=None):
    """Load Input nodes from columns of a CSV file.

    Parameters
    ----------
    filepath : str
        path of CSV file to be loaded.
    exclude_cols : list of str (default: [])
        any columns that should not be loaded as Input.
    eager : bool
        whether to load data as well, making for eager execution.
    nrows : int
        number of rows to load when eager is True. Default is for all rows to load.
    """
    exclude_cols = set(exclude_cols)
    if eager:
        with open(filepath) as f:
            data = pd.read_csv(f, nrows=nrows)
            cols = data.columns
            nodes = [InputNode(col)(data[col].values) for col in cols if not col in exclude_cols]
    else:
        with open(filepath) as f:
            cols = pd.read_csv(f, nrows=2).columns
            nodes = [InputNode(col) for col in cols if not col in exclude_cols]
    return {node.name: node for node in nodes}


def from_sql(connection, query, eager=False, nrows=None):
    """Load Input nodes from columns of a Pandas dataframe.

    Parameters
    ----------
    connection : Connection
        database connection to load from.
    query : str
        query to execute in connection to load columns.
    eager : bool
        whether to load data as well, making for eager execution.
    nrows : int
        number of rows to load when eager is True. Default is for all rows to load.
    """
    col_query = query + ' LIMIT 2'
    cursor = connection.execute(col_query)
    cols = [col[0] for col in cursor.description]
    nodes = [InputNode(col) for col in cols]

    if eager:
        if nrows:
            query += ' LIMIT {}'.format(nrows)
        data = np.array(cursor.execute(query).fetchall()).T
        nodes = [node(datacol) for node, datacol in zip(nodes, data)]

    return {node.name: node for node in nodes}



from .generator import *


def PandasData(dataframe, exclude_cols=[], nrows=None):
    """Load fixed data from Pandas Dataframe into Megatron Input nodes, one for each column.

    Parameters
    ----------
    dataframe : Pandas.DataFrame
        the dataframe to be used.
    exclude_cols : list of str (default: [])
        any columns that should not be loaded as Input.
    nrows : int (default: None)
        number of rows to load. If None, loads all rows.
    """
    return next(PandasGenerator(dataframe, nrows, exclude_cols))


def CSVData(filepath, exclude_cols=[], nrows=None):
    """Load fixed data from CSV filepath into Megatron Input nodes, one for each column.

    Parameters
    ----------
    filepath : str
        the CSV filepath to be loaded from.
    exclude_cols : list of str (default: [])
        any columns that should not be loaded as Input.
    nrows : int (default: None)
        number of rows to load. If None, load all rows.
    """
    return next(CSVGenerator(filepath, nrows, exclude_cols))


def SQLData(connection, query):
    """Load fixed data from SQL query into Megatron Input nodes, one for each column.

    Parameters
    ----------
    connection : Connection
        a database connection to any valid SQL database engine.
    query : str
        a valid SQL query according to the engine being used, that extracts the data for Inputs.
    """
    return next(SQLGenerator(connection, query, -1))



from .. import utils


class Node:
    """Base class of pipeline nodes.

    Parameters
    ----------
    inbound_nodes : list of megatron.Node
        nodes who are to be connected as inputs to this node.

    Attributes
    ----------
    inbound_nodes : list of megatron.Node
        nodes who are to be connected as inputs to this node.
    outbound_nodes : list of megatron.Node
        nodes to whom this node is connected as an input.
    output : np.ndarray
        holds the data output by the node's having been run on its inputs.
    outbounds_run : int
        number of outbound nodes that have been executed.
        this is a helper for efficiently removing unneeded data.
    """
    def __init__(self, inbound_nodes):
        self.inbound_nodes = inbound_nodes
        self.outbound_nodes = []
        self.output = None
        self.outbounds_run = 0
        self.is_output = False
        self.is_eager = False

    def traverse(self, *path):
        """Return a Node from elsewhere in the graph by navigating to it from this Node.

        A negative number indicates moving up to a parent, a positive number down to a child.
        The number itself is a 1-based index into the parents/children, from left to right.
        For example, a step of -2 will go to the second parent, while a step of 3 will go to
        the third child.

        Parameters
        ----------
        path : *ints
            Arbitrary number of integers indicating the steps in the path.

        Returns
        -------
        Node
            the node at the end of the provided path.
        """
        node = self
        for step in path:
            if step < 0:
                node = node.inbound_nodes[-step-1]
            elif step > 0:
                node = node.outbound_nodes[step-1]
        return node


class InputNode(Node):
    """A pipeline node holding input data as a Numpy array.

    It is always an initial node in a Pipeline (has no inbound nodes) and, when run,
    stores its given data (either from a feed dict or a function call) in its output.

    Parameters
    ----------
    name : str
        a name to associate with the data; the keys of the Pipeline feed dict will be these names.
    shape : tuple of int
        the shape, not including the observation dimension (1st), of the Numpy arrays to be input.

    Attributes
    ----------
    name : str
        a name to associate with the data; the keys of the Pipeline feed dict will be these names.
    shape : tuple of int
        the shape, not including the observation dimension (1st), of the Numpy arrays to be input.
    """
    def __init__(self, name, shape=()):
        self.name = name
        self.shape = shape
        super().__init__([])

    def load(self, observations):
        """Validate and store the data passed in.

        Parameters
        ----------
        observations : np.ndarray
            data from either the feed dict or the function call, to be validated.

        Raises
        ------
        megatron.utils.ShapeError
            error indicating that the shape of the data does not match the shape of the node.
        """
        self.validate_input(observations)
        self.output = observations

    def validate_input(self, observations):
        """Ensure shape of data passed in aligns with shape of the node.

        Parameters
        ----------
        observations : np.ndarray
            data from either the feed dict or the function call, to be validated.

        Raises
        ------
        megatron.utils.ShapeError
            error indicating that the shape of the data does not match the shape of the node.
        """
        if hasattr(observations, 'shape') and (list(observations.shape[1:]) != list(self.shape)):
            raise utils.errors.ShapeError(self.name, self.shape, observations.shape[1:])

    def __call__(self, observations):
        """Run the node, storing the given data. Intended for eager execution.

        Parameters
        ----------
        observations : np.ndarray
            data from either the feed dict or the function call, to be validated.

        Raises
        ------
        megatron.utils.ShapeError
            error indicating that the shape of the data does not match the shape of the node.
        """
        self.load(observations)
        self.is_eager = True
        return self


class TransformationNode(Node):
    """A pipeline node holding a Transformation.

    It connects to a set of input Nodes (of class Node or Input) and, when run,
    applies its given Transformation, storing the result in its output variable.

    Parameters
    ----------
    layer : megatron.Layer
        the Layer to be applied to the data from its inbound Nodes.
    inbound_nodes : list of megatron.Node / megatron.Input
        the Nodes to be connected to this node as input.
    layer_out_index : int (default: 0)
        when a Layer has multiple return values, shows which one corresponds to this node.

    Attributes
    ----------
    transformation : megatron.Transformation
        the transformation to be applied to the data from its input Nodes.
    output : None or np.ndarray
        is None until Node is run; when run, the Numpy array produced is stored here.
    is_fitted : bool
        indicates whether the Transformation inside the Node
        has, if necessary, been fit to data.
    """
    def __init__(self, layer, inbound_nodes, layer_out_index=0):
        super().__init__(inbound_nodes)
        self.layer = layer
        self.layer_out_index = layer_out_index
        self.num_path_outbounds = None
        self.name = self.layer.name

    def _clear_inbounds(self):
        for in_node in self.inbound_nodes:
            if (not in_node.is_output) and in_node.outbounds_run == len(in_node.outbound_nodes):
                in_node.output = None

    def partial_fit(self):
        """Apply partial fit method from Layer to inbound Nodes' data."""
        inputs = [node.output for node in self.inbound_nodes]
        self.layer.partial_fit(*inputs)

    def fit(self):
        """Apply fit method from Layer to inbound Nodes' data."""
        inputs = [node.output for node in self.inbound_nodes]
        try:
            self.layer.fit(*inputs)
        except Exception:
            print("Error thrown by layer named {}".format(self.layer.name))
            raise

    def transform(self, prune=True):
        """Apply and store result of transform method from Layer on inbound Nodes' data.

        Parameters
        ----------
        prune : bool (default: True)
            whether to erase data from intermediate nodes after they are fully used.
        """
        inputs = [node.output for node in self.inbound_nodes]
        try:
            self.output = utils.generic.listify(self.layer.transform(*inputs))[self.layer_out_index]
        except Exception:
            print("Error thrown by layer named {}".format(self.layer.name))
            raise

        if any(in_node.is_eager for in_node in self.inbound_nodes):
            self.is_eager = True
        elif prune:
            for in_node in self.inbound_nodes:
                in_node.outbounds_run += 1
            self._clear_inbounds()



from collections import defaultdict
from .generic import listify


def topsort(output_nodes):
    """Returns the path to the desired Transout_typeionNode through the Pipeline.

    Parameters
    ----------
    output_node : Transout_typeionNode
        the target terminal node of the path.

    Returns
    -------
    list of Transout_typeionNode
        the path from input to output that arrives at the output_node.
    """
    output_nodes = listify(output_nodes)
    visited = defaultdict(int)
    order = []

    def dfs(node):
        visited[node] += 1
        for in_node in node.inbound_nodes:
            if in_node not in visited:
                dfs(in_node)
        if visited[node] <= 1:
            order.append(node)

    for output_node in output_nodes:
        dfs(output_node)
    return order



from .. import utils
from .core import Layer
from ..nodes.auxiliary import KerasNode


class Sklearn(Layer):
    def __init__(self, sklearn_transformation):
        super().__init__()
        self.transformation = sklearn_transformation
        self.name = self.transformation.__class__.__name__

    @property
    def metadata(self):
        return self.transformation.__dict__

    def partial_fit(self, *inputs):
        if hasattr(self.transformation, 'partial_fit'):
            self.transformation.partial_fit(*inputs)
        else:
            msg = "Layer {} does not support partial_fit"
            raise NotImplementedError(msg.format(self.transformation.__class__.__name__))

    def fit(self, *inputs):
        self.transformation.fit(*inputs)

    def transform(self, *inputs):
        if hasattr(self.transformation, 'transform'):
            return self.transformation.transform(*inputs)
        else:
            # don't use the labels for this
            return self.transformation.predict(inputs[0])


class Keras(Layer):
    def __init__(self, keras_model):
        super().__init__(n_outputs=len(keras_model.outputs))
        self.model = keras_model
        self.n_inputs = len(self.model.inputs)
        self.name = 'KerasModel'

    def _call(self, inbound_nodes, name=None):
        if name is None and self.n_outputs > 1:
            name = [None for i in range(self.n_outputs)]
        elif len(utils.generic.listify(name)) != self.n_outputs:
            raise ValueError("Number of names does not match number of outputs")

        if self.n_outputs > 1:
            if any(node.output is not None for node in inbound_nodes):
                raise Exception("Keras nodes cannot be run in eager mode")
            out_nodes = [KerasNode(self, inbound_nodes, i) for i in range(self.n_outputs)]
            for node in inbound_nodes:
                node.outbound_nodes += out_nodes
            out = {node.name: node for node in out_nodes}
        else:
            if any(node.output is not None for node in inbound_nodes):
                raise Exception("Keras nodes cannot be run in eager mode")
            out_node = KerasNode(self, inbound_nodes)
            for node in inbound_nodes:
                node.outbound_nodes.append(out_node)
            out = out_node
        return out

    def _split_inputs(self, inputs):
        if self.n_inputs > 1 and self.n_outputs > 1:
            X = inputs[:self.n_inputs]
            Y = inputs[self.n_inputs:]
        elif self.n_inputs > 1:
            X = inputs[:-1]
            Y = inputs[-1]
        elif self.n_outputs > 1:
            X = inputs[0]
            Y = inputs[1:]
        else:
            X, Y = inputs
        return X, Y

    def partial_fit(self, *inputs):
        X, Y = self._split_inputs(list(inputs))
        self.model.train_on_batch(X, Y)

    def fit(self, *inputs, epochs):
        X, Y = self._split_inputs(list(inputs))
        self.model.fit(X, Y, epochs=epochs)

    def fit_generator(self, generator, steps_per_epoch, epochs):
        def _split_generator(generator):
            for inputs in generator:
                X, Y = self._split_inputs(inputs)
                yield X, Y
        self.model.fit_generator(_split_generator(generator),
                                 steps_per_epoch=steps_per_epoch, epochs=epochs)

    def transform(self, *inputs):
        # don't use the labels for this
        if self.n_inputs == 1:
            preds = self.model.predict(inputs[0])
        else:
            preds = self.model.predict(list(inputs[:self.n_inputs]))
        return preds



import numpy as np
import pandas as pd
from .core import StatelessLayer, StatefulLayer


class Cast(StatelessLayer):
    """Re-defines the data type for a Numpy array's contents.

    Parameters
    ----------
    new_type : type
        the new type for the array to be cast to.
    """
    def __init__(self, new_type):
        super().__init__(new_type=new_type)

    def transform(self, X):
        return X.astype(self.kwargs['new_type'])


class AddDim(StatelessLayer):
    """Add a dimension to an array.

    Parameters
    ----------
    axis : int
        the axis along which to place the new dimension.
    """
    def __init__(self, axis):
        super().__init__(axis=axis)

    def transform(self, X):
        return np.expand_dims(X, self.kwargs['axis'])


class OneHotRange(StatefulLayer):
    """One-hot encode a numeric array where the values are a sequence."""
    def partial_fit(self, X):
        if self.metadata:
            self.metadata['min_val'] = min([self.metadata['min_val'], X.min()])
            self.metadata['max_val'] = max([self.metadata['max_val'], X.max()])
        else:
            self.metadata['min_val'] = X.min()
            self.metadata['max_val'] = X.max()

    def transform(self, X):
        return (np.arange(self.metadata['min_val'], self.metadata['max_val']+1) == X[..., None]) * 1


class OneHotLabels(StatefulLayer):
    """One-hot encode an array of categorical values, or non-consecutive numeric values."""
    def partial_fit(self, X):
        if self.metadata:
            self.metadata['categories'] = np.append(self.metadata['categories'], np.unique(X))
            self.metadata['categories'] = np.unique(self.metadata['categories'])
        else:
            self.metadata['categories'] = np.unique(X)

    def transform(self, X):
        return (self.metadata['categories'] == X[..., None]) * 1


class Reshape(StatelessLayer):
    """Reshape an array to a given new shape.

    Parameters
    ----------
    new_shape : tuple of int
        desired new shape for array.
    """
    def __init__(self, new_shape):
        super().__init__(new_shape=new_shape)

    def transform(self, X):
        return np.reshape(X, self.kwargs['new_shape'])


class Flatten(StatelessLayer):
    """Reshape an array to be 1D."""
    def transform(self, X):
        return X.flatten()


class SplitDict(StatelessLayer):
    def __init__(self, fields):
        super().__init__(n_outputs=len(fields), fields=fields)

    def transform(self, dicts):
        out = []
        as_df = pd.DataFrame(dicts.tolist())
        for key in self.kwargs['fields']:
            out.append(as_df[key].values)
        return out


class TimeSeries(StatefulLayer):
    """Adds a time dimension to a dataset by rolling a window over the data.

    Parameters
    ----------
    window_size : int
        length of the window; number of timesteps in the time series.
    time_axis : int
        on which axis in the array to place the time dimension.
    reverse : bool (default: False)
        if True, oldest data is first; if False, newest data is first.
    """
    def __init__(self, window_size, time_axis=1, reverse=False):
        super().__init__(window_size=window_size, time_axis=time_axis, reverse=reverse)

    def partial_fit(self, X):
        self.metadata['shape'] = list(X.shape[1:])
        self.metadata['previous'] = np.zeros([self.kwargs['window_size']-1] + self.metadata['shape'])

    def transform(self, X):
        internal = [np.roll(X, i, axis=0) for i in range(self.kwargs['window_size'])]
        internal = np.moveaxis(np.stack(internal), 0, self.kwargs['time_axis'])
        internal = internal[self.kwargs['window_size']:]
        begin = np.concatenate([X[:self.kwargs['window_size']], self.metadata['previous']])
        begin = [np.roll(begin, i, axis=0) for i in range(self.kwargs['window_size'])]
        begin = np.moveaxis(np.stack(begin), 0, self.kwargs['time_axis'])
        begin = begin[:self.kwargs['window_size']]

        self.metadata['previous'] = X[-self.kwargs['window_size']:]
        return np.concatenate([begin, internal])


class Concatenate(StatelessLayer):
    """Combine Nodes along a given axis. Does not create a new axis.

    Parameters
    ----------
    axis : int (default: -1)
        axis along which to concatenate arrays. -1 means the last axis.
    """
    def __init__(self, axis=-1):
        super().__init__(axis=axis)

    def transform(self, *arrays):
        arrays = list(arrays)
        for i, a in enumerate(arrays):
            if len(a.shape) == 1:
                arrays[i] = np.expand_dims(a, -1)
        if self.kwargs['axis'] == -1:
            return np.concatenate(arrays, axis=-1)
        else:
            return np.concatenate(arrays, axis=self.kwargs['axis']+1)


class Slice(StatelessLayer):
    """Apply Numpy array slicing. An arbitrary number of exclusive slices can be used.

    Slices (passed as hyperparameters) are constructed by the following procedure:
    - To slice from 0 to N: provide the integer N as the slice.
    - To slice from N to the end: provide a 1-tuple of the integer N, e.g. (5,).
    - To slice from M to N exclusive: provide a 2-tuple of the integers M and N, e.g. (3, 6).
    - To slice from M to N with skip P: provide a 3-tuple of the integers M, N, and P.

    Parameters
    ----------
    *slices : int(s) or tuple(s)
        the slices to be applied. Must not overlap. Formatting discussed above.
    """
    def __init__(self, *slices):
        super().__init__(slices=slices)

    def transform(self, X):
        new_slices = []
        for i, s in enumerate(self.kwargs['slices']):
            if s is None:
                new_slices.append(slice(None))
            elif isinstance(s, int):
                new_slices.append(s)
            elif len(s) == 1:
                new_slices.append(slice(s[0], None))
            else:
                new_slices.append(slice(*s))
        return X[tuple(new_slices)]


class Filter(StatelessLayer):
    """Apply given mask to given array along the observation axis to filter out observations."""
    def transform(self, X, mask):
        return X[mask.astype(bool)]



def apply(layer, nodes, **layer_kwargs):
    """Create an instance of the Layer for each given Node with the same hyperparameters.

    The difference between map and apply is that map shares the Layer between Nodes,
    whereas apply creates separate Layers for each Node.

    Parameters
    ----------
    layer : megatron.Layer
        the Layer to apply to the Nodes.
    nodes : list of megatron.Node
        the Nodes to which the Layer should be applied.
    **layer_kwargs : kwargs
        the kwargs to be passed to the initialization of the Layer.
    """
    if isinstance(nodes, dict):
        return {key: layer(**layer_kwargs)(node) for key, node in nodes.items()}
    elif isinstance(nodes, list):
        return [layer(**layer_kwargs)(node) for node in nodes]
    else:
        raise ValueError("Can only apply a layer to multiple nodes")



import warnings


class ShapeError(Exception):
    def __init__(self, name, input_dims, expected_dims):
        msg = "Data fed into '{}' has {} dims; should be {}-D array".format(
            name, input_dims, expected_dims)
        super().__init__(msg)


class DisconnectedError(Exception):
    def __init__(self, missing_inputs):
        base_msg = "The following inputs are not connected to your provided outputs: {}"
        msg = base_msg.format(', '.join([node.name for node in missing_inputs]))
        super().__init__(msg)


def ExtraInputsWarning(extra_inputs):
    msg = "Some input nodes provided that aren't used in the graph: {}"
    msg.format(', '.join([node.node_name for node in extra_inputs]))
    warnings.warn(msg)



import numpy as np
import pandas as pd
from .core import StatelessLayer


class Impute(StatelessLayer):
    """Replace instances of one data item with another, such as missing or NaN with zero.

    Parameters
    ----------
    imputation_dict : dict
        keys of the dictionary are targets to be replaced; values are corresponding replacements.
    """
    def __init__(self, imputation_dict):
        super().__init__(imputation_dict=imputation_dict)

    def transform(self, X):
        for old, new in self.kwargs['imputation_dict'].items():
            if np.isnan(old):
                X[pd.isnull(X)] = new
            else:
                X[X==old] = new
        return X



# This module credits heavy inspiration to a similar Keras module
# https://github.com/keras-team/keras/blob/master/keras/utils/vis_utils.py

import os
from .utils.generic import listify
from IPython.display import SVG
from .nodes import InputNode, MetricNode, ExploreNode
from .layers import Lambda

# check for pydot
try:
    import pydot
except ImportError:
    pydot = None


def _check_pydot():
    """Raise errors if `pydot` or PipelineViz are not properly installed."""
    if pydot is None:
        raise ImportError('Failed to import `pydot`. Please install `pydot` in your '
                          'current environment.')
    try:
        pydot.Dot.create(pydot.Dot())
    except OSError:
        raise OSError('PipelineViz must be installed with its executables included in the $PATH.')


def pipeline_to_dot(pipeline, rankdir='TB'):
    """Convert a megatron Pipeline to dot format for visualization.

    Parameters
    ----------
    pipeline : megatron.Pipeline
        Feature pipeline defined as a pipeline.
    rankdir : str ['TB' or 'LR']
        Direction of pipeline to plot (top to bottom or left to right).

    Returns
    -------
    pydot.Dot
        Dot representation of the Pipeline.
    """
    _check_pydot()
    dot = pydot.Dot()
    dot.set('rankdir', rankdir)
    dot.set('concentrate', True)
    dot.set_node_defaults(shape='record')

    # add nodes
    for node in pipeline.nodes:
        node_id = str(id(node))
        label = node.name
        # make input nodes green, output nodes blue
        if isinstance(node, InputNode):
            color = '#b7e3ff'
        elif node in pipeline.outputs:
            color = '#aeffad'
        elif isinstance(node, MetricNode):
            color = '#ffba54'
        elif isinstance(node, ExploreNode):
            color = '#e6ef34'
        else:
            color = '#e8e8e8'
        pydot_node = pydot.Node(node_id, label=label, style='filled', fillcolor=color)
        dot.add_node(pydot_node)

        # create edges
        for input_node in reversed(node.inbound_nodes):
            input_node_id = str(id(input_node))
            dot.add_edge(pydot.Edge(input_node_id, node_id))

    return dot


def pipeline_imshow(pipeline, rankdir='TB'):
    """Create visualization of pipeline within Jupyter Notebook.

    Parameters
    ----------
    pipeline : megatron.Pipeline
        Feature pipeline defined as a pipeline.
    rankdir : str ['TB' or 'LR']
        Direction of pipeline to plot (top to bottom or left to right).

    Returns
    -------
    IPython.display.SVG
        Display of pipeline.
    """
    dot = pipeline_to_dot(pipeline, rankdir)
    return SVG(dot.create(prog='dot', format='svg'))


def pipeline_imsave(pipeline, save_path='pipeline.png', rankdir='TB'):
    """Save visualization of pipeline to an image file.

    Parameters
    ----------
    pipeline : megatron.Pipeline
        Feature pipeline defined as a pipeline.
    save_path : str
        Specify where to save the pipeline visualization.
    rankdir : str ['TB' or 'LR']
        Direction of pipeline to plot (top to bottom or left to right).
    """
    dot = pipeline_to_dot(pipeline, rankdir)
    _, extension = os.path.splitext(save_path)
    if not extension:
        extension = 'png'
    else:
        extension = extension[1:]
    dot.write(save_path, format=extension)



import math
import numpy as np
import matplotlib.pyplot as plt
from ..nodes.auxiliary import ExploreNode
from .. import utils


class Explorer:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _call(self, nodes, name):
        """Produce an ExploreNode acting on the given inbound nodes.

        Parameters
        ----------
        nodes : list of megatron.Node
            nodes to be passed to the explorer function.
        name : str
            name to be associated with this ExploreNode.
        """
        nodes = utils.generic.listify(nodes)
        out_node = ExploreNode(self, nodes, name)
        for node in nodes:
            node.outbound_nodes.append(out_node)
        if all(node.output is not None for node in nodes):
            out_node.explore()
        return out_node

    def __call__(self, nodes, name):
        return self._call(nodes, name)

    def evaluate(self, *inputs):
        """Run exploration function on given input data."""
        raise NotImplementedError

    def __setitem__(self, key, value):
        self.kwargs[key] = value


class Describe(Explorer):
    """A statistical summary of the given array, similar to Pandas dataframes."""
    def __init__(self, axis=0):
        super().__init__(axis=axis)

    def explore(self, X):
        return {
            'mean': np.mean(X, axis=self.kwargs['axis']),
            'sd': np.std(X, axis=self.kwargs['axis']),
            'min': np.min(X, axis=self.kwargs['axis']),
            '25%': np.percentile(X, 25, axis=self.kwargs['axis']),
            '50%': np.percentile(X, 50, axis=self.kwargs['axis']),
            '75%': np.percentile(X, 75, axis=self.kwargs['axis']),
            'max': np.max(X, axis=self.kwargs['axis'])
        }


class Histogram(Explorer):
    def __init__(self, **plot_kwargs):
        super().__init__(**plot_kwargs)

    def explore(self, X):
        if len(X.shape) > 2:
            raise ValueError("Data has too many dimensions")
        elif len(X.shape) == 1:
            fig, ax = plt.subplots()
            ax.hist(X, **self.kwargs)
            return fig
        else:
            fig = plt.figure()
            nrows = int(math.sqrt(X.shape[1]))
            ncols = math.ceil(X.shape[1] / nrows)
            for i in range(X.shape[1]):
                ax = fig.add_subplot(nrows, ncols, i+1)
                ax.hist(X[:, i], **self.kwargs)
            fig.tight_layout()
            return fig


class Scatter(Explorer):
    def __init__(self, **plot_kwargs):
        super().__init__(**plot_kwargs)

    def explore(self, X, Y):
        if len(X.shape) > 1 or len(Y.shape) > 1:
            raise ValueError("Arrays must be one-dimensional")
        fig, ax = plt.subplots()
        ax.scatter(X, Y, **self.kwargs)
        return fig


class Correlate(Explorer):
    def explore(self, *inputs):
        if len(inputs) == 1:
            return np.corrcoef(inputs[0], rowvar=False)
        else:
            X = np.stack(inputs, axis=1)
            return np.corrcoef(X, rowvar=False)



import os
import inspect
import sqlite3
import numpy as np
import pandas as pd
import dill as pickle
from collections import defaultdict
from . import utils
from . import io
from .nodes.core import InputNode, TransformationNode
from .nodes.auxiliary import MetricNode, ExploreNode, KerasNode
from .layertools import wrappers


class Pipeline:
    """Holds the core computation graph that maps out Layers and manipulates data.

    Parameters
    ----------
    inputs : list of megatron.Node(s)
        input nodes of the Pipeline, where raw data is fed in.
    outputs : list of megatron.Node(s)
        output nodes of the Pipeline, the processed features.
    name : str
        unique identifying name of the Pipeline.
    version : str
        version tag for Pipeline's cache table in the database.
    storage_db : Connection (defeault: 'sqlite')
        database connection to be used for input and output data storage.

    Attributes
    ----------
    inputs : list of megatron.Node(s)
        input nodes of the Pipeline, where raw data is fed in.
    outputs : list of megatron.Node(s)
        output nodes of the Pipeline, the processed features.
    path : list of megatron.Nodes
        full topological sort of Pipeline from inputs to outputs.
    nodes : dict of list of megatron.Node(s)
        separation of Nodes by type.
    eager : bool
        when True, TransformationNode outputs are to be calculated on creation. This is indicated by
        data being passed to an InputNode node as a function call.
    name : str
        unique identifying name of the Pipeline.
    version : str
        version tag for Pipeline's cache table in the database.
    storage: Connection (defeault: None)
        storage database for input and output data.
    """
    def __init__(self, inputs, outputs, metrics=[], explorers=[],
                 name=None, version=None, storage=None, overwrite=False):
        self.eager = False
        self.inputs = utils.flatten(utils.listify(inputs))
        self.outputs = utils.flatten(utils.listify(outputs))
        self.metrics = utils.flatten(utils.listify(metrics))
        self.explorers = utils.flatten(utils.listify(explorers))

        self.path = utils.pipeline.topsort(self.outputs)
        self.metric_path = utils.pipeline.topsort(self.metrics)
        self.explore_path = utils.pipeline.topsort(self.explorers)
        self.nodes = list(set(self.path + self.metric_path + self.explore_path))

        # wipe output data and learned metadata in case of eager execution
        for node in self.nodes:
            node.output = None
            if hasattr(node, 'metadata'):
                node.metadata = {}

        # identify output nodes
        for node in self.outputs:
            node.is_output = True

        # remove edges that aren't in path
        for node in self.nodes:
            node.inbound_nodes = [in_node for in_node in node.inbound_nodes
                                  if in_node in self.nodes]
            node.outbound_nodes = [out_node for out_node in node.outbound_nodes
                                   if out_node in self.nodes]

        # ensure input data matches with input nodes
        missing_inputs = (set(self.nodes).intersection(self.inputs) - set(self.inputs))
        if len(missing_inputs) > 0:
            raise utils.errors.DisconnectedError(missing_inputs)
        extra_inputs = set(self.nodes).intersection(self.inputs) - set(self.nodes)
        if len(extra_inputs) > 0:
            utils.errors.ExtraInputsWarning(extra_inputs)

        # setup output data storage
        self.name = name
        self.version = version
        if self.version:
            version = str(self.version).replace('.', '_')
        if storage:
            self.storage = io.storage.DataStore(self.name, version, storage, overwrite)
        else:
            self.storage = None

    def _reset_run(self, nodes=None):
        nodes = nodes if nodes else self.nodes
        for node in nodes:
            node.outbounds_run = 0

    def _load_inputs(self, input_data, nodes=None):
        # load data into its corresponding InputNodes
        if nodes is None:
            nodes = self.inputs
        for node in nodes:
            node.load(input_data[node.name])

    def _fit_generator_node(self, node, input_generator, steps_per_epoch, epochs):
        # fit a single node that is not a Keras model to a generator
        path_nodes = utils.pipeline.topsort(node)[:-1]
        input_nodes = [node for node in path_nodes if isinstance(node, InputNode)]
        transform_nodes = [node for node in path_nodes if isinstance(node, TransformationNode)]
        for i, batch in enumerate(input_generator):
            self._load_inputs(batch, input_nodes)
            for parent_node in transform_nodes:
                parent_node.transform(prune=False)
            node.partial_fit()
            if i == (steps_per_epoch * epochs): break

    def _fit_generator_keras(self, node, input_generator, steps_per_epoch, epochs):
        # fit a single node that is a Keras model to a generator
        def _generator(node, input_generator):
            path_nodes = utils.pipeline.topsort(node)[:-1]
            input_nodes = [node for node in path_nodes if isinstance(node, InputNode)]
            transform_nodes = [node for node in path_nodes if isinstance(node, TransformationNode)]
            out_nodes = node.inbound_nodes
            while True:
                for batch in input_generator:
                    self._load_inputs(batch, input_nodes)
                    for node in transform_nodes:
                        node.transform()
                    yield [node.output for node in out_nodes]

        node.fit_generator(_generator(node, input_generator),
                           steps_per_epoch=steps_per_epoch, epochs=epochs)

    def partial_fit(self, input_data):
        """Fit to input data in an incremental way if possible.

        Parameters
        ----------
        input_data : dict of Numpy array
            the input data to be passed to InputNodes to begin execution.
        """
        self._load_inputs(input_data)
        for node in self.path:
            if not isinstance(node, TransformationNode): continue
            node.partial_fit()
            node.transform()
        self._reset_run()

    def fit(self, input_data, epochs=1):
        """Fit to input data and overwrite the metadata.

        Parameters
        ----------
        input_data : 2-tuple of dict of Numpy array, Numpy array
            the input data to be passed to InputNodes to begin execution, and the index.
        epochs : int (default: 1)
            number of passes to perform over the data.
        """
        self._load_inputs(input_data)
        for node in self.path:
            if not isinstance(node, TransformationNode): continue
            if isinstance(node, KerasNode):
                node.fit(epochs=epochs)
            else:
                node.fit()
            node.transform()
        self._reset_run()

    def fit_generator(self, input_generator, steps_per_epoch, epochs=1):
        """Fit to generator of input data batches. Execute partial_fit to each batch.

        Parameters
        ----------
        input_generator : generator of 2-tuple of dict of Numpy array and Numpy array
            generator that produces features and labels for each batch of data.
        steps_per_epoch : int
            number of batches that are considered one full epoch.
        epochs : int (default: 1)
            number of passes to perform over the data.
        """
        if sum(isinstance(node, KerasNode) for node in self.path) > 1:
            raise ValueError("Multiple Keras nodes cannot be present when fitting to generator")
        for node in self.path:
            if not isinstance(node, TransformationNode): continue
            if isinstance(node, KerasNode):
                self._fit_generator_keras(node, input_generator, steps_per_epoch, epochs)
            else:
                self._fit_generator_node(node, input_generator, steps_per_epoch, epochs)

    def transform(self, input_data, index_field=None, prune=True):
        """Execute the graph with some input data, get the output nodes' data.

        Parameters
        ----------
        input_data : dict of Numpy array
            the input data to be passed to InputNodes to begin execution.
        index_field : str
            name of key from input_data to be used as index for storage and lookup.
        keep_data : bool
            whether to keep data in non-output nodes after execution.
            activating this flag can be useful for debugging.
        """
        if index_field:
            index = input_data.pop(index_field)
            if len(index.shape) > 1:
                raise ValueError("Index field cannot be multi-dimensional array; must be 1D")
        else:
            nrows = input_data[list(input_data)[0]].shape[0]
            index = pd.RangeIndex(stop=nrows)

        self._load_inputs(input_data)

        # run transformation nodes to end of path
        for node in self.path:
            if not isinstance(node, TransformationNode): continue
            node.transform(prune=prune)
        if prune:
            self._reset_run()

        output_data = [node.output for node in self.outputs]
        if self.storage:
            self.storage.write(output_data, index)
        return output_data

    def transform_generator(self, input_generator, steps, index=None):
        """Execute the graph with some input data from a generator, create generator.

        Parameters
        ----------
        input_generator : dict of Numpy array
            generator producing input data to be passed to Input nodes.
        steps : int
            number of batches to pull from input_generator before terminating.
        """
        for i, batch in enumerate(input_generator):
            if i == steps: StopIteration()
            yield self.transform(batch, index, prune=False)
        self._reset_run()

    def evaluate(self, input_data, prune=True):
        """Execute the metric Nodes in the Pipeline and get their results.

        Parameters
        ----------
        input_data : dict of Numpy array
            the input data to be passed to InputNodes to begin execution.
        """
        self._load_inputs(input_data)
        for node in self.metric_path:
            if isinstance(node, TransformationNode):
                node.transform(prune=prune)
            elif isinstance(node, MetricNode):
                node.evaluate()
        self._reset_run()
        return {node.name: node.output
                for node in self.metric_path if isinstance(node, MetricNode)}

    def evaluate_generator(self, input_generator, steps):
        """Execute the metric Nodes in the Pipeline for each batch in a generator."""
        for i, batch in enumerate(input_generator):
            if i == steps: StopIteration()
            yield self.evaluate(batch, prune=False)

    def explore(self, input_data, prune=True):
        self._load_inputs(input_data)
        for node in self.explore_path:
            if isinstance(node, TransformationNode):
                node.transform(prune=prune)
            elif isinstance(node, ExploreNode):
                node.explore()
        self._reset_run()
        return {node.name: node.output
                for node in self.explore_path if isinstance(node, ExploreNode)}

    def explore_generator(self, input_generator, steps):
        """Execute the explorer Nodes in the Pipeline for each batch in a generator."""
        for i, batch in enumerate(input_generator):
            if i == steps: StopIteration()
            yield self.explore(batch, prune=False)

    def save(self, save_dir):
        """Store the Pipeline and its learned metadata without the outputs on disk.

        The filename will be {name of the pipeline}{version}.pkl.

        Parameters
        ----------
        save_dir : str
            the desired location of the stored nodes, without the filename.
        """
        # store ref to data outside of Pipeline and remove ref to data in TransformationNodes
        data = {}
        for node in self.nodes:
            data[node] = node.output
            node.output = None
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        # write to disk
        with open('{}/{}{}.pkl'.format(save_dir, self.name, self.version), 'wb') as f:
            # keep same cache_dir too for new pipeline when loaded
            pipeline_info = {'inputs': self.inputs, 'outputs': self.outputs,
                             'path': self.path, 'metric_path': self.metric_path,
                             'explore_path': self.explore_path,
                             'metrics': self.metrics, 'explorers': self.explorers,
                             'name': self.name, 'version': self.version}
            if self.storage:
                storage_info = {'output_names': self.storage.output_names,
                                'dtypes': self.storage.dtypes,
                                'original_shapes': self.storage.original_shapes}
                pipeline_info.update(storage_info)
            pickle.dump(pipeline_info, f)
        # reinsert data into Pipeline
        for node in self.nodes:
            node.output = data[node]


def load_pipeline(filepath, storage_db=None):
    """Load a set of nodes from a given file, stored previously with Pipeline.save().

    Parameters
    ----------
    filepath : str
        the file from which to load a Pipeline.
    storage_db : Connection (default: sqlite3.connect('megatron_default.db'))
        database connection object to query for cached data from the Pipeline.
    """
    with open(filepath, 'rb') as f:
        stored = pickle.load(f)
    P = Pipeline(stored['inputs'], stored['outputs'], stored['metrics'], stored['explorers'],
                 stored['name'], stored['version'], storage_db)
    if storage_db:
        # storage members that were calculated during writing
        P.storage.output_names = stored['output_names']
        P.storage.dtypes = stored['dtypes']
        P.storage.original_shapes = stored['original_shapes']
    return P



import inspect
import functools
import collections
import types


def initializer(func):
    # https://stackoverflow.com/questions/1389180/automatically-initialize-instance-variables.
    sig = inspect.signature(func)

    @functools.wraps(func)
    def wrapper(self, *args, **kargs):
        parameters = [sig.parameters[k] for k in sig.parameters]
        for parameter, arg in zip(parameters[1:len(args)+1], args):
            setattr(self, parameter.name, arg)
        for parameter in parameters[len(args)+1:]:
            if parameter.name in kargs:
                setattr(self, parameter.name, kargs[parameter.name])
            else:
                setattr(self, parameter.name, parameter.default)
        func(self, *args, **kargs)

    return wrapper


def listify(x):
    if isinstance(x, list):
        return x
    elif isinstance(x, dict):
        return list(x.values())
    elif isinstance(x, set):
        return list(x)
    else:
        return [x]


def delistify(x):
    return x[0] if isinstance(x, list) else x


def isinstance_str(obj, classname):
    return classname in [t.__name__ for t in obj.__class__.__mro__[:-1]]


def flatten(L):
    def _flatten(L):
        for l in L:
            if isinstance(l, (list, set, tuple)):
                yield from flatten(l)
            elif isinstance(l, dict):
                yield from flatten(list(l.values()))
            else:
                yield l
    return list(_flatten(L))



import numpy as np
from .core import StatelessLayer

try:
    from nltk.corpus import stopwords
except ImportError:
    pass


class RemoveStopwords(StatelessLayer):
    """Remove common, low-information words from all elements of text array.

    Parameters
    ----------
    language : str (default: english)
        the language in which the text is written.
    """
    def __init__(self, language='english'):
        super().__init__(language=language)

    def transform(self, X):
        stops = set(stopwords.words(self.kwargs['language']))
        return ' '.join(word for word in X.split(' ') if word not in stops)


# function will not apply natively to an array of strings; must vectorize
RemoveStopwords.transform = np.vectorize(RemoveStopwords.transform)



from .core import Node, TransformationNode


class MetricNode(Node):
    """A Node that holds an evaluation metric.

    These are not run like other nodes. They do not factor into fit and transform.
    They are only used when the Pipeline's evaluate() method is invoked.

    Parameters
    ----------
    metric : function
        the metric function to be applied to the inbound data.
    inbound_nodes : list of megatron.Node
        the nodes whose data are passed to the metric function.
    name : str
        the name of the node. This is used to refer to the result in the output dict.

    Attributes
    ----------
    metric : function
        the metric function to be applied to the inbound data.
    inbound_nodes : list of megatron.Node
        the nodes whose data are passed to the metric function.
    name : str
        the name of the node. This is used to refer to the result in the output dict.
    """
    def __init__(self, layer, inbound_nodes, name):
        self.layer = layer
        self.inbound_nodes = inbound_nodes
        self.outbound_nodes = []
        self.name = name

    def evaluate(self):
        """Apply metric function to inbound data and store result."""
        inputs = [node.output for node in self.inbound_nodes]
        self.output = self.layer.evaluate(*inputs)


class ExploreNode(Node):
    """A node that holds an exploratory tool, such as a visualization or statistical summary.

    These are not run like other nodes. They do not factor into fit and transform.
    They are only used when the Pipeline's explore() method is invoked.

    Parameters
    ----------
    explorer : function
        the function to be applied to the inbound data that will produce the analysis.
    inbound_nodes : list of megatron.Node
        the nodes whose data are passed to the exploration function.
    name : str
        the name of the node. This is used to refer to the result in the output dict.
    """
    def __init__(self, layer, inbound_nodes, name):
        self.layer = layer
        self.inbound_nodes = inbound_nodes
        self.outbound_nodes = []
        self.name = name

    def explore(self):
        """Apply explorer to inbound data and store result."""
        inputs = [node.output for node in self.inbound_nodes]
        self.output = self.layer.explore(*inputs)


class KerasNode(TransformationNode):
    """A particular TransformationNode that holds a Keras Layer."""
    def partial_fit(self):
        inputs = [node.output for node in self.inbound_nodes]
        self.layer.fit(*inputs)
        self._clear_inbounds()

    def fit(self, epochs=1):
        inputs = [node.output for node in self.inbound_nodes]
        self.layer.fit(*inputs, epochs=epochs)
        self._clear_inbounds()

    def fit_generator(self, generator, steps_per_epoch, epochs=1):
        """Execute Keras model's fit_generator method.

        Parameters
        ----------
        generator : generator
            data generator to be fit to. Should yield tuples of (observations, labels).
        steps_per_epoch : int
            number of batches that are considered one full epoch.
        epochs : int
            number of epochs to run for.
        """
        self.layer.fit_generator(generator, steps_per_epoch=steps_per_epoch, epochs=epochs)



import inspect
from ..nodes import InputNode, TransformationNode
from .. import utils


class Input(InputNode):
    """Wrapper for Input nodes to make them appear as a Layer, for consistency."""
    pass


class Layer:
    """Base class of all layers.

    Parameters
    ----------
    n_outputs (default: 1)
        number of distinct data, and thus nodes, output by the layer's transform.
    **kwargs
        hyperparameters of the transformation function.

    Attributes
    ----------
    n_outputs
        number of distinct data, and thus nodes, output by the layer's transform.
    kwargs
        hyperparameters of the transformation function.
    """
    def __init__(self, n_outputs=1, **kwargs):
        self.n_outputs = n_outputs
        self.kwargs = kwargs
        self.name = self.__class__.__name__

    def _call(self, nodes):
        """Creates a TransformationNode associated with this Layer and the given InputNode(s).

        When running eagerly, will perform a fit and transform.

        Parameters
        ----------
        nodes : list of megatron.InputNode / megatron.TransformationNode
            the input nodes, whose data are to be passed to transform_fn when run.
        """
        nodes = utils.generic.listify(nodes)

        if self.n_outputs > 1:
            out_nodes = [TransformationNode(self, nodes, i)
                        for i in range(self.n_outputs)]
            for node in nodes:
                node.outbound_nodes += out_nodes
            if all(node.output is not None for node in nodes):
                for out_node in out_nodes:
                    out_node.fit()
                    out_node.transform()
            out = out_nodes
        else:
            out_node = TransformationNode(self, nodes)
            for node in nodes:
                node.outbound_nodes.append(out_node)
            if all(node.output is not None for node in nodes):
                out_node.fit()
                out_node.transform()
            out = out_node
        return out

    def __call__(self, nodes):
        return self._call(nodes)

    def partial_fit(self, *inputs):
        """Update metadata based on given data in an iterative fashion.

        Parameters
        ----------
        inputs : numpy.ndarray(s)
            the input data to be fit to; could be one array or a list of arrays.
        """
        pass

    def fit(self, *inputs):
        """Overwrite metadata based on given data.

        Parameters
        ----------
        inputs : numpy.ndarray(s)
            the input data to be fit to; could be one array or a list of arrays.
        """
        pass

    def transform(self, *inputs):
        """Apply transformation to given input data.

        Parameters
        ----------
        inputs : np.ndarray(s)
            input data to be transformed; could be one array or a list of arrays.
        """
        raise NotImplementedError


class StatelessLayer(Layer):
    """A layer holding a stateless transformation."""
    pass


class StatefulLayer(Layer):
    """A layer holding a stateful transformation.

    For custom functions that are stateful, and thus require to be fit,
    writing a Transformation subclass is required rather than using a Lambda wrapper.

    Parameters
    ----------
    n_outputs (default: 1)
        number of distinct data, and thus nodes, output by the layer's transform.
    **kwargs : dict
        the hyperparameters of the transformation function.

    Attributes
    ----------
    n_outputs (default: 1)
        number of distinct data, and thus nodes, output by the layer's transform.
    kwargs : dict
        the hyperparameters of the transformation function.
    metadata : dict
        stores any necessary metadata, which is defined by child class.
    """
    def __init__(self, n_outputs=1, **kwargs):
        super().__init__(n_outputs=n_outputs, **kwargs)
        self.metadata = {}

    def partial_fit(self, *inputs):
        """Update metadata based on given batch of data or full dataset.

        Contains the main logic of fitting. This is what should be overwritten by all child classes.

        Parameters
        ----------
        inputs : numpy.ndarray(s)
            the input data to be fit to; could be one array or a list of arrays.
        """
        msg = "Layer {} does not support partial_fit() or it has not been defined yet"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def fit(self, *inputs):
        """Overwrite metadata based on given batch of data or full dataset.

        Parameters
        ----------
        inputs : numpy.ndarray(s)
            the input data to be fit to; could be one array or a list of arrays.
        """
        self.metadata = {}
        self.partial_fit(*inputs)


class Lambda(StatelessLayer):
    """A layer holding a stateless transformation.

    For custom functions that are stateless, and thus do not require to be fit,
    a Lambda wrapper is preferred to creating a Transformation subclass.

    Parameters
    ----------
    transform_fn : function
        the function to be applied, which accepts one or more
        Numpy arrays as positional arguments.
    n_outputs (default: 1)
        number of distinct data, and thus nodes, output by the layer's transform.
    **kwargs
        keyword arguments to whatever custom function is passed in as transform_fn.

    Attributes
    ----------
    transform_fn : function
        the function to be applied, which accepts one or more
        Numpy arrays as positional arguments.
    n_outputs (default: 1)
        number of distinct data, and thus nodes, output by the layer's transform.
    kwargs
        keyword arguments to whatever custom function is passed in as transform_fn.
    """
    def __init__(self, transform_fn, n_outputs=1, **kwargs):
        super().__init__(n_outputs=n_outputs, **kwargs)
        self.transform_fn = transform_fn
        if self.transform_fn.__name__ == '<lambda>':
            self.name = 'Lambda: anon'
        else:
            self.name = 'Lambda: {}'.format(self.transform_fn.__name__)

    def transform(self, *inputs):
        """Apply associated function to given input data.

        Parameters
        ----------
        inputs : np.ndarray(s)
            input data to be passed to transform_fn; could be one array or a list of arrays.
        """
        return self.transform_fn(*inputs, **self.kwargs)



from setuptools import setup, find_packages

long_description = '''
Megatron is a framework for building computation graphs for
feature engineering in machine learning, with Numpy arrays as the data type.
Use Megatron if you want to:
    - Do feature engineering in a modular and functional way, building up features one step at a time
    - Use disk space to save time by caching feature sets for easy reloading
    - Train feature engineering modules on training data and apply them to testing data
    - Write custom functions for complex transformations, but access built-in functions for quick and common transformations
    - Build feature engineering like you build Keras models (the API is heavily inspired by Keras)
Or any combination of these.
Megatron is distributed under the MIT license.
'''

setup(name='Megatron',
      version='0.5.1',
      description='A computation graph library for feature engineering with Numpy data',
      long_description=long_description,
      author='Nash Taylor',
      author_email='nashtaylor22@gmail.com',
      url='https://github.com/ntaylorwss/megatron',
      download_url='https://github.com/ntaylorwss/megatron/archive/master.zip',
      license='MIT',
      packages=find_packages(),
      install_requires=[
        'numpy',
        'pandas',
        'dill',
      ],
      classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
      ])



import hashlib
import inspect
import numpy as np
from .generic import isinstance_str


def hash_path(nodes):
    h = hashlib.md5()
    for node in nodes:
        to_str = []
        if isinstance_str(node, 'InputNode'):
            to_str.append(node.output)
        else:
            if isinstance_str(node.layer, 'Lambda'):
                to_str.append(inspect.getsource(node.layer.transform_fn))
            elif isinstance_str(node.layer, 'StatefulLayer'):
                to_str.append(inspect.getsource(node.layer.partial_fit))
                to_str.append(inspect.getsource(node.layer.fit))
                to_str.append(inspect.getsource(node.layer.transform))
                to_str += list(node.kwargs.values())
                for metadata in node.layer.metadata.values():
                    to_str.append(metadata)
            elif isinstance_str(node.layer, 'StatelessLayer'):
                to_str.append(inspect.getsource(node.layer.transform))
                to_str += list(node.layer.kwargs.values())

        for s in to_str:
            if isinstance_str(s, 'ndarray'):
                h.update(str(bytes(s)).encode())
            else:
                h.update(str(s).encode())
    return h.hexdigest()



import random
import numpy as np
import copy


class _mapped_call:
    def __init__(self, call):
        self.call = call

    def __call__(self, nodes):
        if isinstance(nodes, dict):
            return {key: self.call(node) for key, node in nodes.items()}
        elif isinstance(nodes, list):
            return [self.call(node) for node in nodes]
        else:
            raise ValueError("Layer can only be mapped when provided multiple nodes")


def map(layer):
    """Cause a layer's transformation to be mapped to each node called on.

    Parameters
    ----------
    layer : megatron.Layer
        Layer object to be wrapped.
    """
    layer._call = _mapped_call(layer._call)
    return layer


class _vectorized_func:
    def __init__(self, func):
        self.func = func

    def __call__(self, X):
        return np.array([self.func(row) for row in X])


def vectorize(layer):
    """Cause a layer's transformation to be mapped to each row in the input data.

    Parameters
    ----------
    layer : megatron.Layer
        Layer object to be wrapped.
    """
    layer.transform = _vectorized_func(layer.transform)
    return layer


class _probabilistic_func:
    def __init__(self, func, p):
        self.func = func
        self.p = p

    def __call__(self, X):
        if random.random() <= self.p:
            return self.func(X)
        else:
            return X


def maybe(layer, p):
    """Cause a layer's transformation to operate probabilistically.

    Layers wrapped this way can operate only on one node.

    Parameters
    ----------
    layer : megatron.Layer
        Layer object to be wrapped.
    p : float
        the probability with which the transformation should be applied.
        on negative trials, the data will simply pass through the layer.
    """
    layer.transform = _probabilistic_func(layer.transform)
    return layer



from .core import Input, StatefulLayer, Lambda
from . import adapters
from . import explore
from . import image
from . import metrics
from . import missing
from . import numeric
from . import shaping
from . import text
from .adapters import *
from .explore import *
from .image import *
from .metrics import *
from .missing import *
from .numeric import *
from .shaping import *
from .text import *
